/*	+-------------------------------------------------------------------------------------------------------------------------------+
	| Interface Compact Flash / PIC + FAT16 avec un PIC18F452																		|																							|
	+-------------------------------------------------------------------------------------------------------------------------------+
	| Copyright (C) 2005 Charles RINCHEVAL (charles.r@digitalspirit.org)															|
	+-------------------------------------------------------------------------------------------------------------------------------+
	| Ce programme est libre, vous pouvez le redistribuer et/ou le modifier selon les termes de la Licence Publique Gnrale GNU 	|
	| publie par la Free Software Foundation (version 2 ou bien toute autre version ultrieure choisie par vous).					|
	| Ce programme est distribu car potentiellement utile, mais SANS AUCUNE GARANTIE, ni explicite ni implicite,					|
	| y compris les garanties de commercialisation ou d'adaptation dans un but spcifique.											|
	| Reportez-vous  la Licence Publique Gnrale GNU pour plus de dtails.														|
	| Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en mme temps que ce programme, si ce n'est pas le cas,	|
	| crivez  la Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, tats-Unis. 					|
	+-------------------------------------------------------------------------------------------------------------------------------+
 */

#include "file.h"


/*	Ouvre un fichier dans le mode spcifi
	@param pathname Le chemin jusqu'au fichier
	@param mode Le mode d'ouverture
	@return Retourne null en cas d'erreur
 */
FILE* fopen(const char* pathname, const rom unsigned char* mode) {

	char fd = -1;		// On limite  128 fichier :)
	char *name;
	unsigned char i;

	DEBUGF("fopen(\"%s\", \"%c\");\n", pathname, *mode);
	
	if (pathname[0] != '/') {
		DEBUGF("'%s' is not an absolute path.\n", (void*)pathname);
		return -1;
	}

	// Si aucun fichier n'est ouvert, on nettoie la structure ...
/*	if (FileNbrOpen == 0) {
		for (i = 0; i < MAX_OPEN_FILES; i++)
			memset(&tOpenFile[i], 0x00, sizeof(FileInfo));
		fd = 0;
	} else {*/
		// ... sinon, on recherche un emplacement libre...
		for (i = 0; i < MAX_OPEN_FILES; i++) {
			if (!(tOpenFile[i].state & FILE_STATE_BUSY)) {
				fd = i;
				break;
			}
		}
//	}

	if (fd == -1) {
		DEBUGF("Too many file open !\n");
		return NULL;
	}

	//	Cod en dr pour le drivers CF
#ifdef WRITE_ENABLED
	bk_load = &cf_buffer_load;
	bk_flush = &cf_buffer_flush;
#endif

	//	Scan des flags
	if (*mode == 'r')
		tOpenFile[fd].flags = O_RDONLY;
	else if (*mode == 'w')
		tOpenFile[fd].flags = O_CREAT | O_WRONLY;
	else if (*mode == 'a')
		tOpenFile[fd].flags = O_CREAT | O_APPEND;

	tOpenFile[fd].state |= FILE_STATE_BUSY;
	FileCurrent = fd;

	// Sparation chemin / nom
	name = strrchr(pathname, '/');
	if (name) {
		*name = 0;

		// On se positionne
		if (fat_cd(pathname) < 0) {
			DEBUGF("Error opening folder '%s' !", pathname);
			return -1;
		}

		*name = '/';
		name++;
	}

	// Le fichier n'existe pas, on le cr ?
	if (fat_open_file(name, &tOpenFile[fd].fatFile, tOpenFile[fd].flags) < 0) {
#ifdef WRITE_ENABLED
		if (tOpenFile[fd].flags & O_CREAT) {
			name--;
			*name = 0;
			fat_cd(pathname);
			*name = '/';
			name++;
			fat_create_entry(name, &tOpenFile[fd].fatFile, 0);
		} else {
			tOpenFile[fd].state = 0;
			FileCurrent = 0;
			fd = -1;
		}
#endif
#ifndef WRITE_ENABLED
		tOpenFile[fd].state = 0;
		FileCurrent = 0;
		fd = -1;
#endif
	}

	// Prparation du fichier
#ifdef WRITE_ENABLED
	if (tOpenFile[fd].flags & O_APPEND) {
		tOpenFile[fd].fatFile.currentByte = tOpenFile[fd].fatFile.size;
		if (tOpenFile[fd].fatFile.startCluster != 0) {
			FileCurrent = -1;
			_set_position(fd, 1);
			FileCurrent = fd;
//			bk_load();
		}
	}
#endif

	return (fd + 1);
}


/*	Lit le fichier courant et charge le buffer de count caractre
	@param buf Le buffer  charger
	@param size La taille d'un lment
	@param num Le nombre d'lments  lire
	@param fd Le descripteur de fichier
	@return Le nombre d'lments lus
 */
int fread(char* buf, size_t size, size_t num, FILE fd) {

	unsigned long i, j;

	fd--;
	_set_position(fd, 0);

	if (tOpenFile[fd].flags != O_RDONLY) {
		DEBUGF("Access mode incorrect !");
		return -1;
	}

	for (i = 0; i < num; i++) {
		for (j = 0; j < size; j++)
			buf[i] = fat_get_next_byte(&(tOpenFile[fd].fatFile));
	}

	return i;
}


/*	Retourne une valeur non null si la fin du fichier  t atteinte
	@param fd Le descripteur de fichier
 */
unsigned char feof(FILE fd) {
	fd--;
	return ((tOpenFile[fd].fatFile.currentByte - 1) == tOpenFile[fd].fatFile.size) ? 1 : 0;
}


/*	Positionne le pointeur au dbut du fichier
	@param fd Le descripteur de fichier
 */
void rewind(FILE fd) {
	fd--;
	tOpenFile[fd].fatFile.currentByte = 0;
	tOpenFile[fd].fatFile.currentCluster = tOpenFile[fd].fatFile.startCluster;
}


/*	Retourne la taille du fichier courant
	@param fd Le descripteur de fichier
 */
unsigned long filesize(FILE fd) {
	return tOpenFile[fd - 1].fatFile.size;
}


#ifdef WRITE_ENABLED

/*	Ecrit le contenu du buffer  la suite du fichier
	@param buf Le buffer  charger
	@param size La taille d'un lment
	@param num Le nombre d'lments  crire
	@param fd Le descripteur de fichier
	@return Le nombre d'lments crit
 */
int fwrite(char* buf, size_t size, size_t num, FILE fd) {

	unsigned int offset = 0;
	unsigned long cmpt, jj, pos,sector = 0;
	size_t i;

	fd--;
	_set_position(fd, 1);

	if (tOpenFile[fd].flags & O_WRONLY &&
		tOpenFile[fd].flags & O_CREAT &&
		tOpenFile[fd].flags & O_APPEND) {
		DEBUGF("Access mode incorrect !");
		return -1;
	}

	// Le fichier s'est dj vu attribuer un cluster ?
	if (tOpenFile[fd].fatFile.startCluster == 0x00) {
		tOpenFile[fd].fatFile.startCluster = fat_get_first_free_cluster();
		tOpenFile[fd].fatFile.currentCluster = tOpenFile[fd].fatFile.startCluster;
	}

	// On crit du dbut ou  partir de la fin ?
/*
	if (tOpenFile[fd].flags & O_APPEND) {
		bk_load();
		offset = tOpenFile[fd].fatFile.size;
	}
*/

	tOpenFile[fd].fatFile.currentByte = offset;

	// Ecriture dans le buffer
	for (jj = 0; jj < num; jj++) {

		pos = offset + size;

/*
		if (pos == BK_BUFFER_SIZE) {

			offset = 0;
			fflush(fd++);
			
			cmpt = tOpenFile[fd].fatFile.currentByte & 1023;
			cmpt++;
			if (cmpt == 1024)
				PORTC = 0;
		}
*/

//		memcpy(&BK_Buffer[pos], buf, size);

		// memcpy
		for (i = 0; i < size; i++) {
			if ((pos + i) == BK_BUFFER_SIZE) {
				offset = 0;
				pos = offset + size;

				fflush(fd++);

				cmpt = tOpenFile[fd].fatFile.currentByte & 1023;
				cmpt++;
				if (cmpt == 1024)
					PORTC = 0;
			}
			BK_Buffer[pos + i] = buf[i];
			tOpenFile[fd].fatFile.currentByte++;
		}


//		tOpenFile[fd].fatFile.currentByte += size;
	}

	tOpenFile[fd].state |= FILE_STATE_FLUSH;

	tOpenFile[fd].fatFile.size += size * num;
}


/*	Ecrit un caractre c dans la resource fd
	@param c Le caractre  crire
	@param fd Le descripteur de fichier
 */
int fputc(char c, FILE fd) {

	unsigned int last_cluster;
	unsigned int offset = 0;
	unsigned long cmpt;

	fd--;
	_set_position(fd, 1);

	if (tOpenFile[fd].flags & O_WRONLY &&
		tOpenFile[fd].flags & O_CREAT &&
		tOpenFile[fd].flags & O_APPEND) {
		DEBUGF("Access mode incorrect !");
		return -1;
	}

	// Le fichier s'est dj vu attribuer un cluster ?
	if (tOpenFile[fd].fatFile.startCluster == 0x00) {
		tOpenFile[fd].fatFile.startCluster = fat_get_first_free_cluster();
		tOpenFile[fd].fatFile.currentCluster = tOpenFile[fd].fatFile.startCluster;
	}

	offset = (tOpenFile[fd].fatFile.currentByte % 512);
/*
	if (offset == 512) {
		offset = 0;
		fflush(fd++);
	}
*/
	
//	cmpt = tOpenFile[fd].fatFile.currentByte & 511;
//	cmpt++;
//	if (cmpt == 512) {
	if (!offset) {

		offset = 0;
		fflush(fd + 1);

		if (!(tOpenFile[fd].fatFile.currentByte % 1024)) {
			// le cluster courant doit venir pointer sur le cluster suivant
			last_cluster = tOpenFile[fd].fatFile.currentCluster;	
			tOpenFile[fd].fatFile.currentCluster = fat_get_first_free_cluster();
			fat_set_cluster(last_cluster, tOpenFile[fd].fatFile.currentCluster);
		}
	}

	tOpenFile[fd].fatFile.currentByte++;
	tOpenFile[fd].fatFile.size++;
	BK_Buffer[offset] = c;

	tOpenFile[fd].state |= FILE_STATE_FLUSH;

	return c;
}


/*	Vide le cache dans le fichier
	@param fd Le descripteur de fichier
 */
unsigned char fflush(FILE fd) {
	
	unsigned char tmp;
	unsigned long sector;

	fd--;

	if (tOpenFile[fd].state & FILE_STATE_FLUSH && FileCurrent == fd) {

		// Tout d'abord, on crit le buffer !

		// Dans quel parti du cluster est-on ?
		tmp = (char)((tOpenFile[fd].fatFile.size - 1) / FAT_Bpb.bytsPerSec);
		
		sector = fat_cluster_2_sector(tOpenFile[fd].fatFile.currentCluster);
		fat_start_addr(sector + ((tmp & 0x01) ? 1 : 0), FAT_MODE_WRITE);
		bk_flush();

		// Ensuite, on crit dans les FATs
		fat_set_cluster(tOpenFile[fd].fatFile.currentCluster, 0xFFFF);

		tOpenFile[fd].state ^= FILE_STATE_FLUSH;
	}

}

#endif


/*	Ferme un fichier proprement
	@param fd Le descripteur de fichier
 */
void fclose(FILE fd) {

#ifdef WRITE_ENABLED
	fflush(fd);
#endif

	fd--;

	DEBUGF("fclose(\"%i\");", (char)fd);

	// On met  jour la directory entry !
	fat_set_directory_entry(&tOpenFile[fd].fatFile);

	memset(&tOpenFile[fd - 1], 0x00, sizeof(tOpenFile));

	tOpenFile[fd].state = 0;
	FileCurrent = 0;
}

/*
int _user_putc(unsigned char t) {

	int i = 0;
	tOpenFile[1].state = 0;
}
*/


/*	Vrifie si le pointeur de la CF est bien  l'endroit demand, si non, elle le repositionne
	@param fd Le descripteur de fichier
	@param mode Spcifie si on est en lecture ou criture...
	@access protected ACCES INTERDIT DE L'EXTERIEUR
 */
void _set_position(FILE fd, unsigned char mode) {

	unsigned int byte = 0;
	unsigned long sector = 0;

	if (FileCurrent != fd) {

		// Repositionnement  l'endroit laiss !
		if (tOpenFile[fd].fatFile.currentByte >= 512) {
			sector = tOpenFile[fd].fatFile.currentByte / 512;
			byte = tOpenFile[fd].fatFile.currentByte - (sector * 512);
		} else {
			byte = tOpenFile[fd].fatFile.currentByte;
		}

		sector += fat_cluster_2_sector(
			(tOpenFile[fd].fatFile.currentCluster) ?
			tOpenFile[fd].fatFile.currentCluster : tOpenFile[fd].fatFile.startCluster);

		fat_start_addr(sector, FAT_MODE_READ);
		fat_skip_byte(byte);

		// Si on tait en append, on recharge le buffer
#ifdef WRITE_ENABLED
		if (mode) {
			bk_load();
		}
#endif

		FileCurrent = fd;
	}
}
